/****************************************************************************
+       TAME 1.0
+
+       assignmap.h
 ****************************************************************************/

#ifndef TAME_ASSIGNMAP_H
#define TAME_ASSIGNMAP_H

#include <map.h>
#include <algo.h>
#include "./var.h"

/*{\Manpage {assignmap} {} {Mappa di assegnazioni di Variabili} {Ass}}*/

class assignmap : public map <var, var> {
/*{\Mdefinition 
Una istanza del tipo di dati |\Mname| rappresenta un insieme di assegnazioni:
|Atomo = Letterale|.}*/

/***** Service Utility Functions *****/

  void
  _print_create_err (ostream & os, const char motiv[13], const var & v) {
    os  << "Invalid Assignment:"
	<< "the Atom "
	<< v << endl
	<< " has a " << motiv
	<< endl;
   return;	
  }	

  pair<var, var>
  make_assign (const var& a, const var & b) {
    if (a.is_negative()) return (make_pair (not(a), not(b)));
    return (make_pair (a, b));
  }	

  void
  substitute (const var & a, const var & l) {
    for (assignmap::iterator it = begin(); it != end();)
      if (((*it).second) == a) {
        var aux = (*it).first;
        erase (it++);
        insert (make_assign (aux, l));
	return;
      }
      else if ( abs ((*it).second) == a) {
        var aux = (*it).first;
        erase (it ++);
        insert (make_assign (aux, not (l)));
	return;
      } 
      else it ++;
      return;
  }

  void
  extend (const var & a, const var & l) {
    insert (make_assign (a, l));
    substitute (a, l);
    return;
  }
		  
						       
/***** End Service Utility Functions *****/

public:
/*{\Mcreation Ass }*/

  assignmap () {}
/*{\Mcreate crea una nuova istanza |\Mvar| del tipo |\Mname| vuota. Resetta le condizioni di errore.}*/ 


/*{\Moperations 1.8 5}*/


  var
  add (const var & Atom, const var & Lit) {
/*{\Mop Inserisce nella lista di assegnazioni, la nuova assegnazione
|Atom = Lit|.
Nel caso |Atom| sia negativo viene cambiato segno ai $2$ letterali.
Restituisce |Atom|.
}*/

#ifdef TAME_DEBUG
    if (Atom.is_costant() ) {
      _print_create_err (cout, "costant value", Atom);
      return FALSE;
    }  
    if (yet_defined (abs(Lit)) ) {
      _print_create_err (cout, "incorrect Ext", Atom);
      return FALSE;
    }
#endif // TAME_DEBUG   

    extend (Atom, Lit);
    return Atom;
  }

  bool
  yet_defined (const var & v) {
/*{\Mop ritorna se la variabile |v| ha gi\'a una definizione in |\Mname|.}*/
    assignmap::iterator it;
    it = find (v);
    if (it != end()) return true;
    return false;
  }  

  void
  operator += (const assignmap & B) {
/*{\Mbinop Assegna a |\Mvar| l'unione insiemistica |\Mvar| $ \cup  B$}*/
    assignmap Aux;
    set_union (begin(), end(), B.begin(), B.end(),
               inserter (Aux, Aux.begin()) );
    *this = Aux;
  }

  assignmap
  operator - (const assignmap & B) {
/*{\Mop Restituisce la differenza $A - B$.}*/
    assignmap Aux;
    set_difference (begin(), end(), B.begin(), B.end(),
                    inserter (Aux, Aux.begin()));
    return Aux;		    
  }

  assignmap
  operator & (const assignmap & B) {
/*{\Mbinop Restituisce l'intersezione insiemistica |\Mvar| $ \cap B$.}*/
    assignmap Aux;
    set_intersection (begin(), end(), B.begin(), B.end(),
                      inserter (Aux, Aux.begin()) );
    return Aux;
  }

  friend
  ostream &
  operator << (ostream & os, const assignmap & Ass) {
/*{\Mbinopfunc Stampa su |os| le assegnazioni contenute in |\Mname|, nel formato:\\
|<Atom1> ::= <Lit1>|\\
\ldots \\
|<Atomn> ::= <Litn>|.}*/
    if (Ass.empty())
      return os << "** Empty Assignment Set **" << endl;
    for (assignmap::const_iterator it = Ass.begin();
         it != Ass.end();
         it ++) 
      os << (*it).first << " ::= " << (*it).second << endl;   
    return os;
  }

/*{\Mtext 
\bigskip
{\bf Funzioni esterne}
\medskip
}*/

friend
bool
store_Ass (assignmap & Ass, assignmap & Bss) {
/*{\Mfunc Sposta tutto il contenuto di |\Mvar| in |Bss|. \\
\precond 	\\
		|\Mvar| $\neq \emptyset$ \\
		|Bss| ottenuta solo per inserimenti successivi di elementi di
		|\Mvar|. \\
Restituisce se \`e stato effettuato lo spostamento.		
}*/
  if (Ass.empty () ) return false;
  for (assignmap::iterator it = Ass.begin ();
       it != Ass.end();
       it ++) {
    Bss.insert (make_pair ( it->first, it->second));
    Ass.erase (it);
  }
  return true;
}  

};
/*{\Mimplementation
Le |\Mname| sono implementate tramite gli standard STL associative container:\\
|template <class Key, class T, less, allocator<T> class map;| dove: \\
|class Key| \`e istanziata al tipo |var| \\
|class T| \`e istanziata al tipo |var| \\
|less| \`e la funzione "standard" di ordinamento delle |var|: 
|bool operator < (const var & a, const var b)|\\
|allocator <var>| \`e l'allocatore predefinito dalla STL. 


\bigskip
Le assegnazioni sono implementate tramite |map <var, var>| in modo che:
\begin {enumerate}
\item il fatto che |Atom| sia anche la chiave della singola assegnazione garamtisce che non si inseriscano $2$ assegnazioni distinte per lo stesso Atomo.
\item La ricerca del valore da sostituire ad ogni ad ogni occorrenza di |Atom|
viene ottiizzata dagli algoritmi interni alla STL, in quanto le |map<Key, T>| sono ordinate in base ai valori della chiave e della funzione |operator< (Key, Key)|.
\item Non \`e stato necessario definire un tipo base:
|class assign : public pair <var, var>| poich\`e questa operazione \`e 
effettuata internamente alla |map <var, var>|.
\end {enumerate}.

\bigskip
{\bf Iterazioni}
\medskip
Sono possibili tutte le iterazioni implementate nelle |map<Key, T>|, in 
particolare:
\begin {itemize}
\item con i tipi |assignmap::iterator| e |assignmap::const|\_|iterator| 
definiti in cicli (i.e. |while|, |for|)
\item con l'operatore |operator []|
\end {itemize}

\bigskip
{\bf Sostituzioni}
\medskip
Qualora un Atomo sia da "ridefinire" sostituendo cio\`e il valore della chiave
con un altro (i.e. quando si effettua una sostituzione, in seguito ad una nuova 
assegnazione), per preservare l'ordinamento della mappa, viene prima eliminata
la assegnazione da modificare e poi reinserita gi\'a aggiornata.
}*/

#endif //TAME_ASSIGNMAP_H  
