/*****************************************************************************
+	TAME 1.0
+
+	rules.h
 *****************************************************************************/

#include "./formula.h"
#include "./xorque.h"

/*{\Manpage {Calculating Rules for Formulas} {F} {Rules}  }*/

/*{\Mdefinition
Qui sono raccolte le funzioni utilizzate per semplificare le formule. Tutte le 
funzioni effettuano |side-effect|.}*/

/*{\Moperations 1.8 5}*/

bool
RedOr (orclause & c);
/*{\Mfunc Riduce la clausola |c|, eliminando le ridondanze, effettuando
"side-effect". Esegue le seguenti azioni:
\begin {itemize}
\item elimina le costanti |TRUE|: $ T \lor C : C \neq \emptyset \rightarrow T $
\item elimina le costanti |FALSE|: $ F \lor C \rightarrow C$
\item elimina le ridondanze di segno opposto: $ L \lor \lnot L \lor C \rightarrow T$
\end {itemize}
Restituisce se la clausola \'e stata ridotta.
Osserviamo esplicitamente che le ridondanze (solo di segno concorde) sono 
eliminate automaticamente poich\`e le clausole sono implementate come |set|.
}*/

bool
RedXor (xorclause & x);
/*{\Mfunc Riduce la clausola |x|, eliminando le ridondanze, effettuando
"side-effect". Esegue le seguenti azioni:
\begin {itemize}
\item elimina le costanti |TRUE|: $ T \oplus L \oplus C \rightarrow \lnot L \oplus C$
\item elimina le costanti |FALSE|: $ F \oplus C \rightarrow C$
\item elimina le ridondanze di segno opposto: $ L \oplus \lnot L \oplus C \rightarrow T \oplus C$
\end {itemize}
Restituisce |true| se la clausola \'e stata ridotta.
Osserviamo esplicitamente che le ridondanze (anche di segno discorde) sono 
eliminate automaticamente poich\`e le clausole |xor| sono implementate come 
|set|.\\
\`E tuttavia necessario che si usi la funzione |add()| ogni qualvolta si voglia inserire un nuovo elemento.
}*/

int
SimpOr (formula & F, const var A, const var B, xorque & XQ);
/*{\Mfunc semplifica l'insieme di clausole |F.C| in base alla assegnazione: 
$A := B$, sostituendo ogni occorrenza di $A$ con $B$.
Restituisce il numero delle clausole semplificate. 
}*/

int
SimpXor (formula & F, const var A, const var B, xorque & XQ);
/*{\Mfunc semplifica l'insiemi di clausole |F.X| in base alla assegnazione:
$A := B$, sostituendo ogni occorrenza di $A$ con $B$.
Restituisce il numero delle clausole semplificate. 
}*/

int
SimpQue (xorque & XQ, const var A, const var B);
/*{\Mfunc semplifica il buffer delle assegnazioni, eseguendo anche la riduzione.
}*/

bool
Simp (formula & F, xorque & XQ);
/*{\Mfunc semplifica |XQ|, |F.C|, |F.X|, |F.S|, in base alle assegnazioni
contenute in |XQ|. Restituisce |true| se non ci sono state contraddizioni,
|false| altrimenti.}*/

#ifdef SIMP_SEL
int
SimpSel (formula & F, const var A, const var B);
/*{\Mfunc semplifica l'insieme di clausole |F.S| in base alla assegnazione:
$A := B$, sostituendo ogni occorrenza di $A$ con $B$.
Restituisce il numero di clausole semplificate.}*/

int
SimpSel (formula & F, const var A, const var B, selectxor::iterator IT);
/*{\Mfunc come la funzione precedente, senza semplificare la selezione puntata
da |IT|.}*/

#endif // SIMP_SEL

bool
UnitOr (orclause & x, xorque & XQ);
/*{\Mfunc Ricerca all'interno di |F.C| ogni clausola contenente un solo 
letterale ($L = \pm A$).
\begin {itemize}
\item se $L = False$ setta il flag di bottom() su F.C e restituisce |false|
\item se $A \neq T$ inserisce $A := \pm True$ nello stack delle assegnazioni 
|F.Ass|, cancella la clausola da |F.C|, restituisce |true|
\item se $L = T$ cancella la clausola da |F.C| e restituisce |true|
\end {itemize}
Restituisce il numero di |unit-clauses| trovate. Nel caso sia trovato un bottom
restituisce $0$.
}*/

bool
UnitXor (xorclause & x, xorque & XQ);
/*{\Mfunc Restituisce se |x| \'e una clausola contenente un solo letterale |L|
\begin {itemize}
\item se $L = False$ setta il flag di bottom() su |XQ| e restituisce |false|
\item se $A \neq T$ inserisce $A := \pm True$ in |XQ| e restituisce |true|
\item se $L = T$ restituisce |true|
\end {itemize}
}*/

int
Or3Xor (formula & F);
/*{\Mfunc Sposta le clausole di $3$ letterali da |F.C| a |F.X|. Ovvero:\\
$\forall c \in$ |F.C| $: c = a \lor b \lor c$ if $\exists c1, c2, c3 \in C$ con
 $c1 = a \lor \neg b \lor \neg c, \
  c2 = \neg a \lor b \lor \neg c, \
  c3 = \neg a \lor \neg b \lor c$ \\
cancella $c, c1, c2, c3$ da |F.C| ed inserisce $x = a \oplus b \oplus c$ in 
|F.X|  
Restituisce il numero di nuove clausole |XOR| trovate.
}*/

var
Select (formula & F);
/*{\Mfunc Seleziona una clausola |x| di |F.X| ed un letterale |A| in essa 
presente per prenderla come definizione di |A|. 
Per eliminare cicli nella riduzione di gauss $\forall B \in x, x$ non ha una 
definizione in |F.S| e |A| $\not = \pm T$.
Restituisce |A| se la ricerca ha avuto successo, |False| altrimenti.
Va usato nelle clausole con almeno $3$ letterali.
}*/

bool
Unselect (formula & F, var A);
/*{\Mfunc deseleziona da |F.S| la definizione di A, se esiste. Ritorna se ha
avuto successo.}*/

int
Gauss (formula & F, var A, xorque & XQ);
/*{\Mfunc Riduce le clausole dell'insieme |F.X|, |F.S| sulla base della |var| 
|A| e della sua definizione |x| in |F.S|.
Calcola il risolvente di Gauss rispetto a |x|, riducendolo e controllando se
sia divenuto una clausola con un unico letterale.
Cancella la premessa di |x|.
Inserisce il risolvente in |x|.
Restituisce il numero di clausole |XOR| ridotte.
}*/

int
EqvOr (formula & F, xorque & XQ);
/*{\Mfunc Sposta le clausole di $2$ letterali da |F.C| ad |XQ|.
Qualora vengano trovate $2$ clausole del tipo:\\
$C1 = \{A, B\}$ e $C2 = \{\lnot A, \lnot B \}$.\\
Cancella le variabili |A| e |B| dall'insieme |F.V| delle proposizioni di |F|.
Restituisce il numero di nuove assegnazioni trovate.
}*/
 
int
EqvXor (formula & F, xorque & XQ);
/*{\Mfunc Sposta le clausole di $2$ letterali da |F.X| ad |XQ|.
Per ogni clausola $X = \{A, B\}$ trovata cancella le variabili |A| e |B| 
dall'insieme |F.V| delle proposizioni di |F|.
Restituisce il numero di nuove assegnazioni trovate.
}*/

var
Split (const formula & F, lassign & VB);
/*{\Mfunc Funzione che impone una condizione su una della variabili della 
formula |F|, effettuando la scelta in base alla function object euristica
|Priority|.
Restituisce:
\begin {itemize}
\item la variabile assegnata, se il calcolo ha avuto esito positivo 
\item la costante |FALSE|, se l'insieme delle clausole |OR| \'e vuoto
\end {itemize}
}*/

int
RedUnit (formula & F, xorque & XQ);
/*{\Mfunc Esegue le riduzioni necessarie sulle clausole di |F| e la ricerca di 
eventuali unit clause, che vengono inserite in |XQ|.}*/
