/****************************************************************************
+       TAME 1.0
+
+       selectxor.h
 ****************************************************************************/

#ifndef TAME_SELECTXOR_H
#define TAME_SELECTXOR_H

#include <map.h>
#include <algo.h>
#include "./xorclause.h"

/*{\Manpage {selectxor} {} {Mappa delle Xorclause selezionate} {S}}*/

class selectxor : public map <var, xorclause> {

/*{\Mdefinition 
Una istanza del tipo di dati |\Mname| rappresenta un insieme di xorclause 
selezionate, ciascuna avente come chiave la variabile di cui costituisce la 
"definizione". 
}*/

  int incorrectSel;

/***** Service Utility Functions *****/

  void
  _set_incorrectSel (int state) {incorrectSel = state;}
 
  pair <var, xorclause>
  make_def (const var & v, const xorclause & x) {
    var temp = abs (v);
    xorclause aux = x;
    xorclause::iterator it;
    if ( (it = aux.find (not (temp)) ) != aux.end()) {
      aux.erase (it);
      return make_pair (temp, aux);
    }
    aux.erase (temp);
    aux = not (aux);
    return make_pair (temp, aux);
  }
      
/***** End Service Utility Functions *****/

public:
/*{\Mcreation 3 }*/

  selectxor () {_set_incorrectSel (0);}
/*{\Mcreate crea una nuova istanza |\Mvar| del tipo |\Mname| vuota.}*/ 


/*{\Moperations 1.8 5}*/
 
  int
  status () {return incorrectSel;}
/*{\Mop restituisce il possibile stato di errore dovuto ad una improvvida
espansione di |\Mvar|:\\
$0$ tutto ok\\
$1$ variabile non presente nella |xorclause| da aggiungere.\\
$2$ |xorclause| troppo corta.\\
$3$ variabile gi\'a definita.}*/

  void
  unset_bottom () {incorrectSel = 0;}
/*{\Mop Resetta la condizione di contraddizione.}*/


  bool
  add (const var & v, const xorclause & x) {
/*{\Mop Inserisce nella mappa di selezione, la nuova xorclause |x|.
che $\forall$ xorclause $y$ : $v,y \not\in$ |\Mvar|.\\
\precond	
		$\pm v \in x$;
		$x.size() \geq 3$;
		$\forall$ xorclause $y$ : $v,y \not\in$ |\Mvar|.

In tutti questi casi vengono settati i flag di errore, da leggere con:
|int S.status ()|.}*/
    xorclause::const_iterator itpos, itneg;
    itpos = x.find (v);
    itneg = x.find ( not (v)); 
    if ((itpos == x.end() ) && (itneg == x.end())) {
      _set_incorrectSel (1);
      return false;
    } 
    if (x.size () < 3) {
      _set_incorrectSel (2);
      return false;
    }  
    var temp = abs (v);		
    if (yet_defined (temp) ) {
      cout << "var: " << temp << "gia' definita in " << *this ;
      _set_incorrectSel (3);
      return false;
    }  
    insert (make_def (v, x));
    return true;
  }

  bool
  yet_defined (const var & v) {
/*{\Mop ritorna se la variabile |v| ha gi\'a una definizione in |\Mname|.}*/
    selectxor::const_iterator it = find (v);
    if (it != end()) return true;
    return false;
  }  

  void
  operator += (const selectxor & D) {
/*{\Mbinopfunc }*/
    set_union (begin(), end(), D.begin(), D.end(),
               inserter (*this, this->begin() ) );
  }

  selectxor
  operator & (const selectxor & D) {
/*{\Mbinopfunc }*/
    selectxor Aux;
    set_intersection (begin(), end(), D.begin(), D.end(),
                      inserter (Aux, Aux.begin() ) );
    return Aux;
  }



  friend
  ostream &
  operator << (ostream & os, const selectxor & S) {
/*{\Mbinopfunc Stampa su |os| le selezioni contenute in |\Mname|, nel formato:\\
|<Var1> ::= <Xorclause1>|\\
\ldots \\
|<Varn> ::= <Xorclausen>|.}*/
    if (S.empty())
      return os << "** Empty Selected Xor Set **" << endl;
    selectxor::const_iterator it;  
    for (it = S.begin(); it != S.end(); it ++) {
      os << (*it).first << " ::= ";
      const xorclause & def = (*it).second;
      os << def << endl;   
    }  
    return os;
  }

};

/*{\Mimplementation
Le |\Mname| sono implementate tramite gli standard STL associative container:\\
|template <class Key, class T, less, allocator<T> class map;| dove: \\
|class Key| \`e istanziata al tipo |var| \\
|class T| \`e istanziata al tipo |xorclause| \\
|less| \`e la funzione "standard" di ordinamento delle |var|:
|bool operator < (const var & a, const var b)|\\
|allocator <var>| \`e l'allocatore predefinito dalla STL.
La variabile da definire viene imagazzinata con il suo valore assoluto\\
la definizione viene immagazzinata previo eliminazione della variabile definita
e con l'opportuna negazione di una dei letterali in essa contenuti.\\
In questo modo \`e possibile eseguire una ricerca pi\'u veloce dei letterali
contenuti in un |xorclause| che possiedono una definizione in |\Mvar|,
ricercandone il valore assoluto.

\bigskip
Si \`e scelto di implementare |map <var, xorclause>| in modo che:
\begin {enumerate}
\item il fatto che |v| sia anche la chiave della singola selection garamtisce che non si inseriscano $2$ assegnazioni distinte per lo stesso Atomo.
\item La ricerca del valore da sostituire ad ogni ad ogni occorrenza di |v|
viene ottiizzata dagli algoritmi interni alla STL, in quanto le |map<Key, T>| 
sono ordinate in base ai valori della chiave e della funzione 
|operator< (Key, Key)|.
Inoltre poich\`e le "definizioni" vengono memorizzate per il valore assoluto di
|v| garantisce di poter fare una sola iterazione nella ricerca della variabile
nella |xorclause| (e non $2$, una per ogni segno!)
\item Non \`e stato necessario definire un tipo base:
|class select : public pair <var, xorcluase>| poich\`e questa operazione \`e
effettuata internamente alla |map <var, xorclause>|.
\end {enumerate}.

\bigskip
{\bf Iterazioni}
\medskip
Sono possibili tutte le iterazioni implementate nelle |map<Key, T>|, in
particolare:
\begin {itemize}
\item con i tipi |selectxor::iterator| e |selectxor::const|\_|iterator| 
definiti in cicli (i.e. |while|, |for|)
\item con l'operatore |operator []|
\end {itemize}.

\bigskip
{\bf Complessit\'a}
\medskip
La funzione |status()| ha complessit\'a $O(1)$; |add()|, |yet_defined()|, 
$O(\log n)$, l'operatore |<<|, $O(n)$.

}*/

#endif //TAME_SELECTXOR_H  
