/*****************************************************************************
+	TAME 1.0
+
+	vset.h
 *****************************************************************************/

#ifndef TAME_VSET_H
#define TAME_VSET_H

#include <map.h>
#include "./var.h"
#include <iostream.h>
#include "./occurrence.h"

#define ZEROOCCURRENCE occurrence (0,0,0,0,0)

/*{\Manpage {vset} {} {Set of Unassigned Boolean Variables} {V} }*/

class vset : public map <var, occurrence> {
/*{\Mdefinition
Una istanza |\Mvar| del tipo di dati |\Mname| rappresenta l'insieme delle
variabili, della logica booleana, con il numero di occorrenze positive e 
negative nelle clausole, ma non ancora assegnate. La sua definizione si \'e 
resa necessaria per semplificarne la ricerca in funzioni come |lookahead()| e 
|split()|.}*/


public:

/*{\Mcreation 3}*/

vset () {}
/*{\Mcreate crea una istanza di |\Mvar| e la inizializza all'insieme vuoto.
}*/

/*{\Moperations 1.8 5}*/

void
addor  (const var a) {
/*{\Mop incrementa il numero di occorrenze della variabile |a|, tenendo conto 
del segno, nell'insieme |\Mvar|.
Nel caso |a| non sia presente, la inserisce con occorrenza unitaria.
Per |Red()|, |Simp()|.}*/
  if (a.is_costant())
    return;
  vset::iterator itv = find (abs(a));
  if (itv == end())
    if ( a.is_negative () ) 
      insert (make_pair (abs (a), occurrence (0,1,0,0,0)));
    else insert (make_pair (a, occurrence (1,0,0,0,0)));
  else {
    if (a.is_negative () ) (*itv).second.incr_cneg();
    else (*itv).second.incr_cpos();
  }  
  return;
}

void
addxor  (const var a) {
/*{\Mop incrementa il numero di occorrenze della variabile |a|, non tenendo 
conto del segno, nell'insieme |XorClause|.
Nel caso |a| non sia presente, la inserisce con occorrenza unitaria.
Per |Red()|, |Simp()|.}*/
  if (a.is_costant())
    return;
  vset::iterator itv = find (abs(a));
  if (itv == end()) {
    insert (make_pair (abs (a), occurrence (0,0,1,0,0)));
  } 
  else {
    (*itv).second.incr_xor();
  }  
  return;
}

void
addsel  (const var a) {
/*{\Mop incrementa il numero di occorrenze della variabile |a|, non tenendo
conto del segno, nell'insieme |Select Xor|.
Nel caso |a| non sia gi\'a presente, la inserisce con occorrenza unitaria.
}*/
  if (a.is_costant())
    return;
  vset::iterator itv = find (abs(a));
  if (itv == end()) {
    insert (make_pair (abs (a), occurrence (0,0,0,1,0)));
  }
  else {
    (*itv).second.incr_sel();
  }
  return;
}

void
addass (const var a) {
/*{\Mop incrementa il numero di occorrenze della variabile |a|, nelle 
assegnazioni.}*/
  if (a.is_costant())
    return;
  vset::iterator itv = find (abs (a));
  if (itv == end()) {
    insert (make_pair (abs (a), occurrence (0,0,0,0,1)));
  }
  else {
    (*itv).second.incr_ass();
  }
  return;
}



void
subor (const var a) {
/*{\Mop decrementa il numero di occorrenze della variabile |a| nell'insieme 
delle orclausole.
Nel caso |a| non sia presente esce.
Nel caso il numero di occorrenze di |a| sia esattamente $1$, cancella |a|.
Per |Red()|, |Simp()|.}*/
  if (a.is_costant())
    return;
  vset::iterator itv = find (abs(a));
  if (itv == end())
    return;
  else {
    if (a.is_negative() ) (*itv).second.decr_cneg();
    else (*itv).second.decr_cpos();
    if ( (*itv).second == ZEROOCCURRENCE);
  }
  return;
}

void
subxor (const var a) {
/*{\Mop decrementa il numero di occorrenze della variabile |a| nell'insieme
delle xorclausole.
Nel caso |a| non sia presente esce.
Nel caso che, dopo il decremento, il numero di occorrenze di |a| sia nullo
sia nelle orclause che nelle xorclause, cancella |a|.
Per |Red()|, |Simp()|.}*/
  if (a.is_costant())
    return;
  vset::iterator itv = find (abs(a));
  if (itv == end())
    return;
  else {
    (*itv).second.decr_xor();
    if ( (*itv).second == ZEROOCCURRENCE);
  }
  return;
}

void
subsel (const var a) {
/*{\Mop decrementa il numero di occorrenze della variabile |a| nell'insieme
delle select xor.
Nel caso |a| non sia presente esce.
}*/
  if (a.is_costant())
    return;
  vset::iterator itv = find (abs(a));
  if (itv == end())
    return;
  else {
    (*itv).second.decr_sel();
    if ( (*itv).second == ZEROOCCURRENCE);
  }
  return;
}

void
subass (const var a) {
/*{\Mop decrementa il numero di occorrenze di |a| nelle assegnazioni.}*/
  if (a.is_costant())
    return;
  vset::iterator itv = find (abs(a));
  if (itv == end())
    return;
  else {
    (*itv).second.decr_ass();
    if ( (*itv).second == ZEROOCCURRENCE);
  }
  return;
}

  
void
del (const var & a) { 
/*{\Mop azzera le occorrenze della variabile |a| dall'insieme |\Mvar|.
}*/
  vset::iterator itv;
  itv = find ( abs (a));
  if (itv != end() ) 
    (*itv).second.reset_occurrences();
}    

void
flush() {
/*{\Mop resetta le occorrenze di ogni variabile.}*/
  vset::iterator itv;
  for (itv = begin(); itv != end(); itv ++)
    (*itv).second.reset_occurrences();
}   

void
copy (const vset & G) {
  vset::const_iterator itg;
  if (empty() ) 
    for (itg = G.begin(); itg != G.end(); itg++)
      insert (make_pair (abs ((*itg).first), 
              occurrence ((*itg).second.get_cpos(),
	                  (*itg).second.get_cneg(),
			  (*itg).second.get_xor(),
			  (*itg).second.get_sel(),
			  (*itg).second.get_ass() )));
  else 
    for (vset::iterator itv = begin(); itv != end(); itv++) 
      if ( (itg = G.find ((*itv).first)) != G.end() ) 
        (*itv).second = (*itg).second;
  return ;	
}

vset &
operator += (const vset & G) {
  vset::iterator itv;
  vset::const_iterator itg;
  for (itg = G.begin(); itg != G.end(); itg ++) {
    if ( (itv = find ( (*itg).first) ) != end() )
      (*itv).second += (*itg).second;
  }
  return (*this);
}

friend
ostream &
operator << (ostream & os, const vset & V) {
/*{\Mbinopfunc stampa le variabili contenute con le loro occorrenze.}*/
  if (V.empty())
    return os << "** Empty Occurrence Set **" << endl;
  os << "Lit\t" <<"orpos\t" <<"orneg\t" <<"xor\t" <<"sel\t" <<"ass" << endl;  
  for (vset::const_iterator itv = V.begin();
       itv != V.end();
       itv ++)
    os << (*itv).first  
       << "\t" << (*itv).second.get_cpos()
       << "\t" << (*itv).second.get_cneg()
       << "\t" << (*itv).second.get_xor()
       << "\t" << (*itv).second.get_sel()
       << "\t" << (*itv).second.get_ass() 
       << endl;
  return os;
}

};

/*{\Mimplementation
I |\Mname| sono implementati come map \<|var|, |occurrence|\>. 
Le operazioni |addor()|, |addxor()| prendono $O(1)$.
Le operazioni |subor()|, |subxor()|, |del()| prendono $O(\log n)$ 
Le operazioni |flush()| prendono $O(n)$
dove $n$ \'e il numero di elementi di |\Mvar|.
}*/
#endif //TAME_VSET_H
